
require 'sketchup'
require 'set'
require 'open-uri'

module SeanRegan
  module AutoDovetails
    $versionNumber = 0.3

    # determines if a given point lines on the same line as points start and stop and is between start and stop.
    # returns 0 if point lies on either start or stop, negative if outside start or stop, or positive if in between
    def AutoDovetails.pointBetweenPoints(point, start, stop)
      #Points a,b,c are collinear (all lie on a single line) if b-a,c-a are parallel, which is true iff the cross product (b-a) x (c-a) is zero
      #Given that points a,b,c are collinear, b lies between a and c iff the scalar product (b-a).(c-b) is positive. (Non-negative, if it's OK for b to be coincident with a or c.)
      crossProduct = AutoDovetails.crossProduct(point - start, stop - start)
      if(crossProduct[0] != 0 || crossProduct[1] != 0 || crossProduct[2] != 0)
        return -1
      end

      if(point == start || point == stop)
        return 0
      end
      return AutoDovetails.dotProduct(point - start, stop - point)
    end

    # dot product of two 3d points
    def AutoDovetails.dotProduct(point1, point2)
      return point1[0]*point2[0] + point1[1]*point2[1] + point1[2]*point2[2]
    end

    # cross product of two 3d points
    def AutoDovetails.crossProduct(point1, point2)
      return Geom::Point3d.new(point1[1]*point2[2] - point2[1]*point1[2],
                                point2[0]*point1[2] - point1[0]*point2[2],
                                point1[0]*point2[1] - point1[1]*point2[0]) 
    end

    # multiply each point of a vector by some multiple
    def AutoDovetails.multiplyVector(multiplier, vector)
      return (Geom::Vector3d.new multiplier * vector[0], multiplier * vector[1], multiplier * vector[2])
    end

    # finds the two endpoints of overlap between line1 and line2
    # both line segments must be on the same line
    def AutoDovetails.findStartStop(line1start, line1end, line2start, line2end)
      if(AutoDovetails.pointBetweenPoints(line1start, line2start, line2end) >= 0 && AutoDovetails.pointBetweenPoints(line1end, line2start, line2end) >= 0)
        return [line1start, line1end]
      elsif(AutoDovetails.pointBetweenPoints(line2start, line1start, line1end) >= 0 && AutoDovetails.pointBetweenPoints(line2end, line1start, line1end) >= 0)
        return [line2start, line2end]
      elsif(AutoDovetails.pointBetweenPoints(line2start, line1start, line1end))
        if(line2end.distance(line1start) <= line2end.distance(line1end))
          return [line2start, line1start]
        else
          return [line2start, line1end]
        end
      else
        if(line1end.distance(line2start) <= line1end.distance(line2end))
          return [line1start, line2start]
        else
          return [line1start, line2end]
        end
      end
    end

    def AutoDovetails.isCommonLineInMiddle(point1, point2, commonLine)
      intersectionPoint = Geom.intersect_line_line([point1,point2], commonLine)
      if(point1.distance_to_line(commonLine) < point2.distance_to_line(commonLine) &&
        intersectionPoint != nil &&
        AutoDovetails.pointBetweenPoints(intersectionPoint, point1, point2) >= 0)
        return true
      end
      return false
    end

    $dlg = UI::WebDialog.new("Auto Dovetails", false, nil, 340, 310, 230, 150, true)
    $currentVersion = nil
    def AutoDovetails.load_dovetail_window     
          pathname = File.expand_path( File.dirname(__FILE__) )
          pathname = File.join( pathname, 'inputForm.html' )
          $dlg.set_file( pathname )
          $dlg.allow_actions_from_host("seanregan.com")
          $dlg.set_background_color("ececec")

          #load the settings from file, or create some defaults of it doesn't exist
          settings = Array.new
          if(File.exist?(File.dirname(__FILE__) +"/settings.txt"))
            File.open(File.dirname(__FILE__) +"/settings.txt", "r") do |f|
              f.each_line do |line|
                settings.push(line.strip)
              end
            end
          else
            settings.push "4"
            settings.push "8"
            if(Sketchup.active_model.options["UnitsOptions"]["LengthUnit"] == 2) 
                settings.push "10"
                settings.push "5"
            else
              settings.push "1/4"
              settings.push "1/8"
            end
            settings.push ""
            settings.push "100"
            settings.push "100"
            settings.push "100"
          end

          #load the settings into the web form via javascript
          
          units = "inches"
          if(Sketchup.active_model.options["UnitsOptions"]["LengthUnit"] == 1) 
            units = "feet"
          elsif(Sketchup.active_model.options["UnitsOptions"]["LengthUnit"] == 2) 
            units = "mm"
          elsif(Sketchup.active_model.options["UnitsOptions"]["LengthUnit"] == 3)
            units = "cm"
          elsif(Sketchup.active_model.options["UnitsOptions"]["LengthUnit"] == 4)
              units = "m"
          end
                    
          $dlg.add_action_callback("preload_values") {|dialog, params|
             js_command = "document.getElementById('numberOfTails').value = '#{settings[0]}';\n"
             js_command += "document.getElementById('slope').value = '#{settings[1]}';\n"
             js_command += "document.getElementById('halfPinWidth').value = '#{settings[2]}';\n"
             js_command += "document.getElementById('pinWidth').value = '#{settings[3]}';\n"
             js_command += "document.getElementById('tailHeight').value = '#{settings[4]}';\n"
             js_command += "document.getElementById('centerTailSizing').value = '#{settings[5]}';\n"
             js_command += "document.getElementById('endTailSizing').value = '#{settings[6]}';\n"
             js_command += "document.getElementById('centerPinSizing').value = '#{settings[7]}';\n"
             js_command += "updateUnits('#{units}');\n"
             dialog.execute_script(js_command)
           }

           #handler for when the user clicks the submit button on the web form
           $dlg.add_action_callback("ruby_messagebox") {|dialog, params|
             File.open(File.dirname(__FILE__) +"/settings.txt", "w") do |f|

                f.puts($dlg.get_element_value("numberOfTails"))
                f.puts($dlg.get_element_value("slope"))
                f.puts($dlg.get_element_value("halfPinWidth"))
                f.puts($dlg.get_element_value("pinWidth"))
                f.puts($dlg.get_element_value("tailHeight"))
                f.puts($dlg.get_element_value("centerTailSizing"))
                f.puts($dlg.get_element_value("endTailSizing"))
                f.puts($dlg.get_element_value("centerPinSizing"))
             end

             # clear out the previously completed faces & tails
             $donePins = Array.new
             $doneTails = Array.new
             $edgeAlerted = false
             $heightAlerted = false
             AutoDovetails.drawDovetails($dlg)
           }

           # load my website
           $dlg.add_action_callback("website") {|dialog, params|
             UI.openURL("http://seanregan.com/sketchup/dovetail")
           }

          $dlg.show

          #check my website to see if an update is available
          if($currentVersion == nil)
            begin
              open("http://seanregan.com/sketchup/dovetail/version.txt") do |f|
                $currentVersion = f.read.strip
                if(Gem::Version.new($currentVersion) > Gem::Version.new(@su_autodovetails_extension.version))
                  result = UI.messagebox("An update for AutoDovetails is available!\nLatest version is v#{$currentVersion}, you are currently using v#{@su_autodovetails_extension.version.to_f}.\n\nWould you like to download the update?", MB_YESNO)
                   if result == IDYES
                     UI.openURL("http://seanregan.com/sketchup/dovetail")
                     $dlg.close
                   end
                end
              end
            rescue SocketError
              $currentVersion = 0
            end
          end

    end

    $donePins = Array.new
    $doneTails = Array.new
    $edgeAlerted = false
    $heightAlerted = false
    # method to draw one set of pins on tails on two neighboring faces
    def AutoDovetails.createDovetail(tailComponent, tailFace, tailNormal, oppositeTailEdge, pinComponent, pinFace, commonPinEdge, commonTailEdge, tailDovetailStartPoint, tailDovetailStopPoint, pinDovetailStartPoint, pinDovetailStopPoint, dovetailLength,defaultTailHeight,tailFaceNormal)

      # grab the current settings from the form field
      units = $dlg.get_element_value("unitsHidden")
      slope = $dlg.get_element_value("slope").to_f
      pinWidth = $dlg.get_element_value("pinWidth").to_l
      numberOfTails = $dlg.get_element_value("numberOfTails").to_f
      numberOfPins = numberOfTails - 1
      numberOfCenterPins = numberOfTails % 2 == 0 ? 1 : 2
      numberOfCenterTails = numberOfTails % 2 == 0 ? 2 : 1
      halfPinWidth = $dlg.get_element_value("halfPinWidth").to_l
      tailHeight = $dlg.get_element_value("tailHeight").to_l
      
      #if we want metric, but the document is set up in inches, convert it

      
      centerPinWidth = $dlg.get_element_value("centerPinSizing").to_f * pinWidth / 100
      endTailSizing = $dlg.get_element_value("endTailSizing").to_f / 100
      centerTailSizing = $dlg.get_element_value("centerTailSizing").to_f / 100

      if(tailHeight <= 0)
        tailHeight = defaultTailHeight
      end
      


      # support box joints by entering 0 for slope, otherwise determine how much difference there will be in size from top to bottom of the tail
      if(slope > 0)
        baseWidthChange = tailHeight / slope
      else
        baseWidthChange = 0
      end

      # determine number of end tails (to help us determine tail sizing)
      if(numberOfTails <= 2)
        numberOfEndTails = 0
      else
        numberOfEndTails = 2
      end

      # do some math to get ourselves the size of each tail
      effectiveNumberOfTails = (numberOfTails - numberOfCenterTails - numberOfEndTails) + (numberOfCenterTails * centerTailSizing) + (numberOfEndTails * endTailSizing)
      tailWidth = dovetailLength
      tailWidth -= 2 * halfPinWidth
      tailWidth -= pinWidth * (numberOfPins - numberOfCenterPins)
      tailWidth -= numberOfCenterPins * centerPinWidth
      tailWidth = tailWidth / effectiveNumberOfTails
      centerTailWidth = tailWidth * centerTailSizing
      endTailWidth = tailWidth * endTailSizing
      
        minimumLength = (0.1 * numberOfTails) + (2 * halfPinWidth) + (pinWidth * ([0, numberOfPins - numberOfCenterPins].max)) + (centerPinWidth * numberOfCenterPins)
        
        # ensure that the edge we're about to draw has enough space to create the tails        
        if((baseWidthChange*2 >= tailWidth || baseWidthChange*2 >= centerTailWidth || (numberOfEndTails > 0 && baseWidthChange*2 >=endTailWidth)  || dovetailLength < minimumLength))
         
          if(!$edgeAlerted)
            UI.messagebox "Some edges could not be dovetailed because their was not enough room on the edge(s) to accomodate your current settings"
            $edgeAlerted = true
          end
          return nil
        end


      # this seems to help fix some normals issues encountered at times.  just make sure we don't try to delete the face twice
      if(!$doneTails.include?(tailFace))
        tailComponent.definition.entities.erase_entities(tailFace)
        $doneTails.push(tailFace)
      end

      # this doesn't seem to help...commmentig out
      #if(!$donePins.include?([pinFace, pinDovetailStartPoint, pinDovetailStopPoint]))
        #pinComponent.definition.entities.erase_entities(pinFace)
      #end

      # http://math.stackexchange.com/questions/83404/finding-a-point-along-a-line-in-three-dimensions-given-two-points
      # determine the direction in which to draw the set of tails & pins
      tailVector = tailDovetailStopPoint.vector_to(tailDovetailStartPoint)
      pinVector = pinDovetailStopPoint.vector_to(pinDovetailStartPoint)
      tailVector.normalize!
      pinVector.normalize!

      lowerPosition = halfPinWidth + baseWidthChange
      upperPosition = halfPinWidth


      edgesToDelete = Array.new
      for i in 1..numberOfTails
        currentTailWidth = nil
        currentPinWidth = nil
        # determine if this is one of the center tails, end tails, or just a normal tail
        if(numberOfTails % 2 == 1 &&  i < ((numberOfTails / 2) + 1) && i > (numberOfTails / 2))
          currentTailWidth = centerTailWidth
        elsif(numberOfTails % 2 == 0 && ((i == (numberOfTails / 2)) || (i == (numberOfTails / 2) + 1)))
          currentTailWidth = centerTailWidth
        elsif(i == 1 || i == numberOfTails)
          currentTailWidth = endTailWidth
        else
          currentTailWidth = tailWidth
        end

        # determine if this is a center pin or a regular pin
        if(numberOfTails % 2 == 1 &&  i < ((numberOfTails / 2) + 1) && i > (numberOfTails / 2) - 1)
          currentPinWidth = centerPinWidth
        elsif(numberOfTails % 2 == 0 && i == (numberOfTails / 2))
          currentPinWidth = centerPinWidth
        else
          currentPinWidth = pinWidth
        end


        # determine the location of all the vertices for the pins / tails
        lowerStartPoint = tailDovetailStartPoint - AutoDovetails.multiplyVector(lowerPosition,tailVector)
        pin1Start = pinDovetailStartPoint - AutoDovetails.multiplyVector(lowerPosition,pinVector)
        lowerPosition += currentTailWidth - 2 * baseWidthChange

        lowerEndPoint = tailDovetailStartPoint - AutoDovetails.multiplyVector(lowerPosition,tailVector)
        pin2Start = pinDovetailStartPoint - AutoDovetails.multiplyVector(lowerPosition,pinVector)
        lowerPosition += currentPinWidth + 2 * baseWidthChange

        upperStartPoint = tailDovetailStartPoint - AutoDovetails.multiplyVector(upperPosition,tailVector)
        pin1End = pinDovetailStartPoint - AutoDovetails.multiplyVector(upperPosition,pinVector)
        upperPosition += currentTailWidth
        upperEndPoint = tailDovetailStartPoint - AutoDovetails.multiplyVector(upperPosition,tailVector)
        pin2End = pinDovetailStartPoint - AutoDovetails.multiplyVector(upperPosition,pinVector)
        upperPosition += currentPinWidth

        upperStartPoint.offset!(AutoDovetails.multiplyVector(tailHeight, tailFaceNormal))
        upperEndPoint.offset!(AutoDovetails.multiplyVector(tailHeight, tailFaceNormal))

        # figure out how deep to make the tail to cover the width of the board
        projectedPoint = lowerStartPoint.project_to_line(oppositeTailEdge.line)
        backSideOffset = lowerStartPoint.vector_to(projectedPoint)

        # move the top of the tails out by however tall the tails are
        pin1End.offset!(AutoDovetails.multiplyVector(tailHeight, tailFaceNormal.transform(tailComponent.transformation).transform(pinComponent.transformation.inverse)))
        pin2End.offset!(AutoDovetails.multiplyVector(tailHeight, tailFaceNormal.transform(tailComponent.transformation).transform(pinComponent.transformation.inverse)))

        if(!$doneTails.include?([commonTailEdge, tailDovetailStartPoint, tailDovetailStopPoint]))
          #componentDefinition1.entities.add_line(lowerStartPoint,lowerEndPoint)
          tailComponent.definition.entities.add_line(upperStartPoint,upperEndPoint).find_faces
          tailComponent.definition.entities.add_line(lowerStartPoint,upperStartPoint).find_faces
          tailComponent.definition.entities.add_line(lowerEndPoint,upperEndPoint).find_faces
          tailComponent.definition.entities.add_line(upperStartPoint,upperStartPoint.offset(backSideOffset)).find_faces
          tailComponent.definition.entities.add_line(upperEndPoint,upperEndPoint.offset(backSideOffset)).find_faces
          tailComponent.definition.entities.add_line(lowerStartPoint,lowerStartPoint.offset(backSideOffset)).find_faces
          tailComponent.definition.entities.add_line(lowerEndPoint,lowerEndPoint.offset(backSideOffset)).find_faces

          tailComponent.definition.entities.add_line(upperStartPoint.offset(backSideOffset),upperEndPoint.offset(backSideOffset)).find_faces
          tailComponent.definition.entities.add_line(lowerStartPoint.offset(backSideOffset),upperStartPoint.offset(backSideOffset)).find_faces
          tailComponent.definition.entities.add_line(lowerEndPoint.offset(backSideOffset),upperEndPoint.offset(backSideOffset)).find_faces

          #we'll want to delete the edges as the base of the tail to clean up the model
          edgesToDelete.push([lowerStartPoint,lowerEndPoint])
          edgesToDelete.push([lowerStartPoint.offset(backSideOffset),lowerEndPoint.offset(backSideOffset)])
        end
        
        if(!$donePins.include?([commonPinEdge, pinDovetailStartPoint, pinDovetailStopPoint]))
          # draw the pin face
          newPinFace = pinComponent.definition.entities.add_face(pin1Start,pin1End, pin2End, pin2Start)
          offSetAmount = AutoDovetails.multiplyVector(backSideOffset.length * -1, newPinFace.normal)
          newPinFace.pushpull(backSideOffset.length * -1, false)
          pinComponent.definition.entities.each do |edgeToDelete|
          #newLine = pinComponent.definition.entities.add_line(pin1Start.offset(offSetAmount), pin2Start.offset(offSetAmount))
          #pinComponent.definition.entities.erase_entities(newLine)
            if(edgeToDelete.typename == 'Edge')
              # delete the edges at the base of the pin
               if((edgeToDelete.vertices[0].position == pin1Start && edgeToDelete.vertices[1].position == pin2Start) ||
                  (edgeToDelete.vertices[1].position == pin1Start && edgeToDelete.vertices[0].position == pin2Start))
                  pinComponent.definition.entities.erase_entities(edgeToDelete)
                  break
                end
            end
          end

          # just redraw a few lines to make sure the faces all get filled in correctly
          pinComponent.definition.entities.add_line(pin2Start, pin2End).find_faces
          pinComponent.definition.entities.add_line(pin1Start, pin1End).find_faces
          pinComponent.definition.entities.add_line(pin2Start.transform(offSetAmount), pin2End.transform(offSetAmount)).find_faces
          pinComponent.definition.entities.add_line(pin1Start.transform(offSetAmount), pin1End.transform(offSetAmount)).find_faces

          # not sure why sometimes the extrude leaves the top or bottom face intact...this cleans it up
          pinComponent.definition.entities.each do |pinFace|
            if(pinFace.typename == 'Face')
              pinFace.edges.each do |edge|
               if((edge.vertices[0].position == pin1End && edge.vertices[1].position == pin2End) ||
                  (edge.vertices[1].position == pin1End && edge.vertices[0].position == pin2End) ||
                  (edge.vertices[0].position == pin1Start.transform(offSetAmount) && edge.vertices[1].position == pin2Start.transform(offSetAmount)) ||
                  (edge.vertices[1].position == pin1Start.transform(offSetAmount) && edge.vertices[0].position == pin2Start.transform(offSetAmount)))
                    if(edge.faces.length == 1)
                      pinComponent.definition.entities.erase_entities(edge)
                    end
                  end
               end
            end
          end
        end
      end

      # run through all the edges at the base of the tails we had marked for deletion...and delete them
        tailComponent.definition.entities.each do |edgeToDelete|
          if(edgeToDelete.typename == 'Edge')

            edgesToDelete.each do |vertices|
               if((edgeToDelete.vertices[0].position == vertices[0] && edgeToDelete.vertices[1].position == vertices[1]) ||
                  (edgeToDelete.vertices[1].position == vertices[0] && edgeToDelete.vertices[0].position == vertices[1]))
                  tailComponent.definition.entities.erase_entities(edgeToDelete)
                  break
                end
          end
        end
      end

      # remember that we're done at this location, in case we get duplicates in case of multiple component instances
      $donePins.push([commonPinEdge, pinDovetailStartPoint, pinDovetailStopPoint])
      $doneTails.push([commonPinEdge, pinDovetailStartPoint, pinDovetailStopPoint])
      
      $doneTails.push([commonTailEdge, tailDovetailStartPoint, tailDovetailStopPoint])
      $donePins.push([commonTailEdge, tailDovetailStartPoint, tailDovetailStopPoint])
    end

    # finds all the locations of the joints which need to be turned into dovetails
    def AutoDovetails.drawDovetails(dlg)
      model = Sketchup.active_model
      selection = model.selection
      status = model.start_operation('Draw Dovetails', false)

      result = nil
      # just do a check to see if this component exists elsewhere. 
      # since the transformations will happen to all instances of the component, this could screw up the model
      selection.each do |selectedObject|
        if(selectedObject.typename == 'ComponentInstance')
         if(selectedObject.definition.instances.length > 1)
           selectedObject.definition.instances.each do |componentInstance|
             if(result.nil? == true && !selection.include?(componentInstance))
               result = UI.messagebox("Some of the selected components appear in other locations in this model.\nProceeding may result in partially dovetailed joints.\nIt is recommended that you select all instances of the components to be dovetailed along with the component(s) to be joined.\n\nWould you still like to proceed?", MB_YESNO)
                if result == IDNO
                  return nil
                else
                  break
                end
             end
           end
         end
        end
      end

      numberOfObjectsSelected = selection.length

      if (numberOfObjectsSelected < 2)
        UI.messagebox "Please select at least two adjacent components to dovetail"
        return nil
      end

      facesToDovetail = Set.new
      selection.each do |component1|
        selection.each do |component2|

          if((component1.typename != 'ComponentInstance' && component1.typename != 'Group') ||
             (component2.typename != 'ComponentInstance' && component2.typename != 'Group'))
            UI.messagebox "Selected objects must be Components or Groups in order to be dovetailed"
            return nil
          end

          # make sure we don't try to dovetail to ourselves
          if(component1 == component2)
            next
          end

          componentDefinition1 = component1.definition
          componentDefinition2 = component2.definition

          componentLocation1 = component1.transformation
          componentLocation2 = component2.transformation

          componentEntities1 = componentDefinition1.entities
          componentEntities2 = componentDefinition2.entities

          # loop through all the selected components and see if there are any joined edges
          componentEntities1.each do |entity|
            if(entity.typename == 'Edge')
              vertex1 = entity.vertices[0];
              vertex2 = entity.vertices[1];
              position1 = vertex1.position.transform!(componentLocation1)
              position2 = vertex2.position.transform!(componentLocation1)

              componentEntities2.each do |entity2|

                if(entity2.typename == 'Edge')
                  fullyMatchedVertices = 0
                  vertex3 = entity2.vertices[0]
                  vertex4 = entity2.vertices[1]
                  position3 = vertex3.position.transform!(componentLocation2)
                  position4 = vertex4.position.transform!(componentLocation2)

                  line1 = [position1, position2]
                  line2 = [position3, position4]
                  # if all the vertices of both line segments are on the same line
                  # and there is at least some overlap between the lines.....I'm sure there's a better way to do this, but it seems to work and I don't want to touch it
                  if(position1.on_line?(line2) && position2.on_line?(line2) && position3.on_line?(line1) && position4.on_line?(line1) &&
                    (
                     (AutoDovetails.pointBetweenPoints(position1, position3, position4) >= 0 && AutoDovetails.pointBetweenPoints(position2, position3, position4) >= 0) ||
                     (AutoDovetails.pointBetweenPoints(position3, position1, position2) >= 0 && AutoDovetails.pointBetweenPoints(position4, position1, position2) >= 0) ||
                     (
                      (AutoDovetails.pointBetweenPoints(position3, position1, position2) > 0 && AutoDovetails.pointBetweenPoints(position1, position3, position4) > 0) ||
                      (AutoDovetails.pointBetweenPoints(position3, position1, position2) > 0 && AutoDovetails.pointBetweenPoints(position2, position3, position4) > 0) ||
                      (AutoDovetails.pointBetweenPoints(position4, position1, position2) > 0 && AutoDovetails.pointBetweenPoints(position1, position3, position4) > 0) ||
                      (AutoDovetails.pointBetweenPoints(position4, position1, position2) > 0 && AutoDovetails.pointBetweenPoints(position2, position3, position4) > 0) ||
                      (AutoDovetails.pointBetweenPoints(position1, position3, position4) > 0 && AutoDovetails.pointBetweenPoints(position3, position1, position2) > 0) ||
                      (AutoDovetails.pointBetweenPoints(position1, position3, position4) > 0 && AutoDovetails.pointBetweenPoints(position4, position1, position2) > 0) ||
                      (AutoDovetails.pointBetweenPoints(position2, position3, position4) > 0 && AutoDovetails.pointBetweenPoints(position3, position1, position2) > 0) ||
                      (AutoDovetails.pointBetweenPoints(position2, position3, position4) > 0 && AutoDovetails.pointBetweenPoints(position4, position1, position2) > 0)
                     )))
                     #ensure this is the long edge of the joint
                     vertex1.edges.each do |edge|
                       if(edge != entity && edge.length <= entity.length)
                         vertex3.edges.each do |edge2|
                          if(edge2 != entity2 && edge2.length <= entity2.length)
                            fullyMatchedVertices = 1
                          end
                        end
                       end
                    end
                  end

                  # if we have found our common joint, now lets identify the faces to dovetail
                  if(fullyMatchedVertices == 1)
                    commonEdge1 = entity
                    commonEdge2 = entity2

                    faces1 = commonEdge1.faces
                    if(faces1.size < 2)
                      next
                    end

                    faces2 = commonEdge2.faces
                    if(faces2.size < 2)
                      next
                    end

                    entity1Edges1 = Array.new
                    entity1Edges2 = Array.new
                    entity2Edges1 = Array.new
                    entity2Edges2 = Array.new

                    # find the opposite edges of the common edge for each component
                    faces1[0].edges.each do |edge|
                      if(edge != commonEdge1)
                        if(edge.line[1].parallel?(commonEdge1.line[1]) && !edge.line[0].on_line?(commonEdge1.line))
                          entity1Edges1.push edge
                        end
                      end
                    end

                    faces1[1].edges.each do |edge|
                      if(edge != commonEdge1)
                        if(edge.line[1].parallel?(commonEdge1.line[1]) && !edge.line[0].on_line?(commonEdge1.line))
                          entity1Edges2.push edge
                        end
                      end
                    end

                    faces2[0].edges.each do |edge|
                      if(edge != commonEdge2)
                        if(edge.line[1].parallel?(commonEdge2.line[1]) && !edge.line[0].on_line?(commonEdge2.line))
                          entity2Edges1.push edge
                        end
                      end
                    end

                    faces2[1].edges.each do |edge|
                      if(edge != commonEdge2)
                        if(edge.line[1].parallel?(commonEdge2.line[1]) && !edge.line[0].on_line?(commonEdge2.line))
                          entity2Edges2.push edge
                        end
                      end
                    end

                    if(entity1Edges1.length == 0 || entity1Edges2.length == 0 || entity2Edges1.length == 0 || entity2Edges2.length == 0)
                      next
                    end

                  # loops through all the opposite edges and see what orientation they're in in relation to the common line
                  # and decide if it can be dovetailed
                  entity1Edges1.each do |entity1Edge1|
                    entity1Edges2.each do |entity1Edge2|
                      entity2Edges1.each do |entity2Edge1|
                        entity2Edges2.each do |entity2Edge2|
                          entity1Edge1Vertex1 = entity1Edge1.vertices[0].position.transform!(componentLocation1)
                          entity1Edge1Vertex2 = entity1Edge1.vertices[1].position.transform!(componentLocation1)
                          entity1Edge2Vertex1 = entity1Edge2.vertices[0].position.transform!(componentLocation1)
                          entity1Edge2Vertex2 = entity1Edge2.vertices[1].position.transform!(componentLocation1)
                          entity2Edge1Vertex1 = entity2Edge1.vertices[0].position.transform!(componentLocation2)
                          entity2Edge1Vertex2 = entity2Edge1.vertices[1].position.transform!(componentLocation2)
                          entity2Edge2Vertex1 = entity2Edge2.vertices[0].position.transform!(componentLocation2)
                          entity2Edge2Vertex2 = entity2Edge2.vertices[1].position.transform!(componentLocation2)

                          commonEdge1Vertex1 = commonEdge1.vertices[0].position.transform!(componentLocation1)
                          commonEdge1Vertex2 = commonEdge1.vertices[1].position.transform!(componentLocation1)
                          commonEdge2Vertex1 = commonEdge2.vertices[0].position.transform!(componentLocation2)
                          commonEdge2Vertex2 = commonEdge2.vertices[1].position.transform!(componentLocation2)

                          startEndPoint = AutoDovetails.findStartStop(commonEdge1Vertex1, commonEdge1Vertex2, commonEdge2Vertex1, commonEdge2Vertex2)
                          dovetailStartPoint = startEndPoint[0]
                          dovetailStopPoint = startEndPoint[1]

                          dovetailLength = dovetailStartPoint.distance(dovetailStopPoint)

                          entity1Edge1Line = [entity1Edge1Vertex1, entity1Edge1Vertex2]
                          entity1Edge2Line = [entity1Edge2Vertex1, entity1Edge2Vertex2]
                          entity2Edge1Line = [entity2Edge1Vertex1, entity2Edge1Vertex2]
                          entity2Edge2Line = [entity2Edge2Vertex1, entity2Edge2Vertex2]

                          pinFace = nil
                          tailFace = nil
                          tailNormal = nil
                          tailComponent = nil
                          pinComponent = nil
                          oppositeTailEdge = nil
                          commonPinEdge = nil
                          commonTailEdge = nil
                          defaultTailHeight = nil
                          # using the common edge of what will be the pin board, if common line is in middle of pin & tail endpoints
                          if(AutoDovetails.isCommonLineInMiddle(entity2Edge1Vertex1,entity1Edge1Vertex1, [commonEdge2Vertex1, commonEdge2Vertex2]))
                            pinFace = faces2[0]
                            tailFace = faces1[1]
                            tailNormal = faces1[0].normal
                            tailComponent = component1
                            pinComponent = component2
                            oppositeTailEdge = entity1Edge2
                            commonTailEdge = commonEdge1
                            commonPinEdge = commonEdge2
                          elsif(AutoDovetails.isCommonLineInMiddle(entity2Edge2Vertex1,entity1Edge1Vertex1, [commonEdge2Vertex1, commonEdge2Vertex2]))
                            pinFace = faces2[1]
                            tailFace = faces1[1]
                            tailNormal = faces1[0].normal
                            tailComponent = component1
                            pinComponent = component2
                            oppositeTailEdge = entity1Edge2
                            commonTailEdge = commonEdge1
                            commonPinEdge = commonEdge2
                          elsif(AutoDovetails.isCommonLineInMiddle(entity2Edge1Vertex1,entity1Edge2Vertex1, [commonEdge2Vertex1, commonEdge2Vertex2]))
                            pinFace = faces2[0]
                            tailFace = faces1[0]
                            tailNormal = faces1[1].normal
                            tailComponent = component1
                            pinComponent = component2
                            oppositeTailEdge = entity1Edge1
                            commonTailEdge = commonEdge1
                            commonPinEdge = commonEdge2
                          elsif(AutoDovetails.isCommonLineInMiddle(entity2Edge2Vertex1,entity1Edge2Vertex1, [commonEdge2Vertex1, commonEdge2Vertex2]))
                            pinFace = faces2[1]
                            tailFace = faces1[0]
                            tailNormal = faces1[1].normal
                            tailComponent = component1
                            pinComponent = component2
                            oppositeTailEdge = entity1Edge1
                            commonTailEdge = commonEdge1
                            commonPinEdge = commonEdge2
                          elsif(AutoDovetails.isCommonLineInMiddle(entity1Edge1Vertex1,entity2Edge1Vertex1, [commonEdge1Vertex1, commonEdge1Vertex2]))
                            pinFace = faces1[0]
                            tailFace = faces2[1]
                            tailNormal = faces2[0].normal
                            tailComponent = component2
                            pinComponent = component1
                            oppositeTailEdge = entity2Edge2
                            commonTailEdge = commonEdge2
                            commonPinEdge = commonEdge1
                          elsif(AutoDovetails.isCommonLineInMiddle(entity1Edge2Vertex1,entity2Edge1Vertex1, [commonEdge1Vertex1, commonEdge1Vertex2]))
                            pinFace = faces1[1]
                            tailFace = faces2[1]
                            tailNormal = faces2[0].normal
                            tailComponent = component2
                            pinComponent = component1
                            oppositeTailEdge = entity2Edge2
                            commonTailEdge = commonEdge2
                            commonPinEdge = commonEdge1
                          elsif(AutoDovetails.isCommonLineInMiddle(entity1Edge1Vertex1,entity2Edge2Vertex1, [commonEdge1Vertex1, commonEdge1Vertex2]))
                            pinFace = faces1[0]
                            tailFace = faces2[0]
                            tailNormal = faces2[1].normal
                            tailComponent = component2
                            pinComponent = component1
                            oppositeTailEdge = entity2Edge1
                            commonTailEdge = commonEdge2
                            commonPinEdge = commonEdge1
                          elsif(AutoDovetails.isCommonLineInMiddle(entity1Edge2Vertex1,entity2Edge2Vertex1, [commonEdge1Vertex1, commonEdge1Vertex2]))
                            pinFace = faces1[1]
                            tailFace = faces2[0]
                            tailNormal = faces2[1].normal
                            tailComponent = component2
                            pinComponent = component1
                            oppositeTailEdge = entity2Edge1
                            commonTailEdge = commonEdge2
                            commonPinEdge = commonEdge1
                          else
                            next
                          end

                          defaultTailHeight = nil
                          pinFace.edges.each do |edge|
                            if(edge.line[1].perpendicular?(commonPinEdge.line[1]))
                              defaultTailHeight = edge.length
                            end
                          end

                          # if for some reason we couldn't determine a pin height, just pick something safe-ish
                          if(!defaultTailHeight)
                            defaultTailHeight = 0.5
                          end

                          tailHeight = $dlg.get_element_value("tailHeight").to_l
                          if(tailHeight > defaultTailHeight && !$heightAlerted)
                            result = UI.messagebox "Warning!\n\nYour tail height appears to be taller than the thickness of the pin board.  This plugin does not handle protruding tails very well in this version.  Some post cleanup may be required.\n\nWould you like to continue?",  MB_YESNO
                            if(result == IDNO)
                              return nil
                            end
                            $heightAlerted = true
                          end
                          
                          # add this set to the list of dovetail ops to be completed
                          facesToDovetail.add([tailComponent, tailFace, tailNormal, oppositeTailEdge, pinComponent, pinFace, commonPinEdge, commonTailEdge,
                                              dovetailStartPoint.transform(tailComponent.transformation.inverse),dovetailStopPoint.transform(tailComponent.transformation.inverse),
                                              dovetailStartPoint.transform(pinComponent.transformation.inverse),dovetailStopPoint.transform(pinComponent.transformation.inverse),
                                              dovetailLength, defaultTailHeight, tailFace.normal])
                          break
                        end
                      end
                    end
                  end
                  end
                end
              end
            end
          end
        end
      end

      if(facesToDovetail.length == 0)
        UI.messagebox("Couldn't find anything to dovetail!\n\nPlease ensure that the selected components share at least one butt joint.", MB_OK)
      end

     #facesToDovetail.each do |item|
     #  facesToDovetail.each do |item2|
     #    if((item2[1] == item[1] && item2[11] < item[11]))
     #      facesToDovetail.delete(item2)
     #    end
     #  end
     #end

      facesToDovetail.each do |item|
        AutoDovetails.createDovetail(item[0],item[1],item[2],item[3],item[4],item[5],item[6],item[7],item[8],item[9],item[10],item[11],item[12], item[13], item[14])
      end
      $doneTails = Array.new
      $donePins = Array.new
      
      status = model.commit_operation
      selection.clear
    end

    unless file_loaded?( __FILE__ )
      plugins_menu = UI.menu("Plugins")

      plugins_menu.add_item("Dovetail") { AutoDovetails.load_dovetail_window }
    end

    file_loaded( __FILE__ )
  end
end
